from wasmer import Type

I32 = Type.I32
I64 = Type.I64
F32 = Type.F32
F64 = Type.F64
V128 = Type.V128
EXTERN_REF = Type.EXTERN_REF
FUNC_REF = Type.FUNC_REF

get_types = lambda x: list(map(lambda x: Type(x).name, x))

class CheckFunction:
    def _baseAssertFuncType(self, first, second, msg=None):
        """ The default assertEqual implementation, not type specific. """
        if not first == second:
            msg = self._formatMessage(msg, "Types Differ")
            raise self.failureException(msg)   
     
    def assertSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """ The implementation of assertEqual. """
        if not (seq1 == seq2):
            msg = self._formatMessage(msg, "Types Differ")
            raise self.failureException(msg)   
               
    def _getAssertFuncType(self, first, second):
        if type(first) is type(second):
            asserter = self._type_equality_funcs.get(type(first))
            if asserter is not None:
                if isinstance(asserter, str):
                    asserter = getattr(self, asserter)
                return asserter
        return self._baseAssertFuncType

    def _assertFuncType(self, first, second, msg):
        first, second = get_types(first), get_types(second)
        assertion_func = self._getAssertFuncType(first, second)
        msg = """\n\nTypes Differ (on {}):\nWebAssembly Function: {}\nUnittest FunctionTypes: {}""".format(msg, first, second)
        assertion_func(first, second, msg=msg)


from wasmer import engine, wasi, Store, Module, Instance
from wasmer_compiler_cranelift import Compiler

class WasiModule:
    """ Complied some code from c, c++ or rust please use me """
    def __init__(self, path):
        store = Store(engine.JIT(Compiler))
        module = Module(store, open(path, 'rb').read())
        wasi_version = wasi.get_version(module, strict=True)
        wasi_env = wasi.StateBuilder('test-program').finalize()
        import_object = wasi_env.generate_import_object(store, wasi_version)
        self.instance = Instance(module, import_object)
        self.exports = self.instance.exports
        
    def run_main(self):
        try:
            # Here we go, let's start main function (the program).
            print(f"\nMain function(output):")
            self.instance.exports._start()
        except RuntimeError as e:
            if not (str(e) == "RuntimeError: WASI exited with code: 0"):   
                msg = f"WASI exited uncleanly: {str(e)}"
                msg = self._formatMessage(msg, "WASI exited")
                raise self.failureException(msg)     


from wasmer import engine, Store, Instance
from wasmer import wat2wasm, wasm2wat, ImportObject
from wasmer import Function, Global, Memory, Table
from wasmer import Value
from wasmer import Type as WType, Module as WASMModule
from wasmer_compiler_cranelift import Compiler
from typing import NamedTuple, List, Type
from .extension import CheckFunction, get_types
# cranelift was chosen as it is has fast compilation times and fast execution times
import unittest
import loguru
import pathlib
import traceback

class Module:
    IMPORT_FLAG = False
    
    def __init__(self, path, wat=False):
        self.path = path
        if wat:
            self._import_wat_module()
        else:
            self._import_module()

    def _import_wat_module(self):
        """ import a WAT module """
        self.store = Store()
        self.wasm_bytes = wat2wasm(open(self.path, 'r').read())
        
    def _import_module(self):
        """ import a WASM module """
        self.store = Store(engine.JIT(Compiler))
        self.wasm_bytes = open(self.path, 'rb').read()    
        
    def register(self, name, namespace):
        """ register and import python functions into the module """
        self.import_object = ImportObject()
        self.import_object.register(
            name,
            namespace
        )
        self._IMPORT_FLAG = True
        
    def start(self):
        """ start module and export functions """
        # WASMModule is 'Module' from the wasmer library
        module = WASMModule(self.store, self.wasm_bytes)
        if self._IMPORT_FLAG:
            instance = Instance(module, self.import_object)
        else:
            instance = Instance(module)
        self.exports = instance.exports
        
class FuncType(NamedTuple):
    """ FuncType: holds the param types and results type. """
    params: List[Type]
    result: List[Type]

class _WasmiteCase(unittest.TestCase, CheckFunction):
    """Wasmite Extension of unittest.TestCase"""
    
    def checkFunctionTypes(self, func, func_type):
        """ Check the static types of the param(s) and return of WebAssembly Function """
        self._assertFuncType(func.type.params, func_type.params, "params")
        self._assertFuncType(func.type.results, func_type.result, "return")

class WasmiteCase(_WasmiteCase):
    pass

def main():
    unittest.main(verbosity=2)


from .__main__ import Module 
from .__main__ import WasmiteCase

from .__main__ import FuncType, Value
from .__main__ import Function, Global, Memory, Table
from .extension import I32, I64, F32, F64, V128, EXTERN_REF, FUNC_REF

from .__main__ import main
from .wasi import WasiModule


from operator import attrgetter
from os import path
from setuptools import setup, find_packages

VERSION = "0.1.6"

def read(fname):
    return open(path.join(path.dirname(__file__), fname)).read()

def from_here(relative_path):
    return path.join(path.dirname(__file__), relative_path)

# source env/bin/activate

# rm -rf build dist wasmite.egg-info
# python setup.py sdist bdist_wheel 
# python -m twine upload --skip-existing dist/*
# python -m twine upload dist/*

setup(
    name="wasmite",
    version=VERSION,
    author="Yusuf Ahmed",
    author_email="yusufahmed172@gmail.com",
    packages=find_packages(exclude=['env', ".vscode", "wasmite.egg-info"]),
    description="Wasmite: Webassembly is the future but now it has a testing toolchain",
    long_description=read('Readme.md'),
    long_description_content_type="text/markdown",
    install_requires=['loguru', "wasmer>=1.0.0-alpha3", "wasmer-compiler-cranelift>=1.0.0-alpha3"],
    url="https://github.com/yusuf8ahmed/Wasmite",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Browsers",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Server",
        "Topic :: Software Development :: Assemblers",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: System :: Monitoring",
    ],
    platforms = 'any',
    keywords = ["wasmer", "wasmite", "wasp", "wasm", "debug", "debugging", "unit",
                "unit testing", "unit tests", "unit testing wasm", "unit tests wasm",
                "debug wasm", "debugging wasm"],
    python_requires='!=3.9, !=2.*',
    zip_safe = False
)